
include get.e
include wildcard.e
include sort.e

sequence classes, units  classes = {} units = {}
global constant
classId = 1, className = 2, classStr = 3,
unitId = 1, unitClassId = 2, fromUnit = 3, toUnit = 4, unitName = 5, unitStr =6,
macroStr = 3 

-------------------------------------------------------------------------------

function trim(sequence s)
	integer i1,i2
	i1 = 1  i2 = 0
	for i = 1 to length(s) do
		if s[i] > 32 then i1=i exit end if
	end for
	for i = length(s) to 1 by -1 do
		if s[i] > 32 then i2=i exit end if
	end for
	return s[i1..i2]
end function

sequence buf  buf = {}
function get_definition(atom fn)
	object tmp
	sequence terms,def
	integer p
	def = {}  terms = {}
	while 1 do
		tmp = gets(fn)
		if atom(tmp) then
			return 0
		end if
		tmp = trim(tmp)
		if length(tmp) then
			if tmp[1] != '#' then -- if not comment
				p = find('#',tmp) -- ? trailing comment
				if p then
					if tmp[p-1] != '\\' then
						tmp = tmp[1..p-1]
					end if
				end if
				buf &= tmp -- add to anything leftover from last read
				if not length(def) then -- if not parsing definition
					p = find('[',buf)
					if p then
						def = buf[1..p-1] buf = buf[p+1..length(buf)]
					end if
				end if
				if find(']',buf) then
					p = match(']'& def,buf)
					if p then -- if end definition
						terms &= buf[1..p-1]
						buf = trim(buf[p+1+length(def)..length(buf)])
						exit
					end if -- leave chrs in buffer and add to it
				else -- this line is done add to definition string get another line
					terms &= " " & buf  buf = {}
				end if
			end if
		end if
	end while
	return {def,terms}
end function

function get_terms(sequence line)
	sequence terms,t,n,v
	integer c,p,l	
	terms = {} t = {} p = 1 l = length(line)
	while p <= l do
		c = line[p]
		if c = '\\' then -- escapes
			p += 1
			if p <= l then
				c = line[p]
				if c = ';' then -- semicolon
					t &= c
				elsif c = '\\' then -- backslash
					t &= c
				elsif c = 'r' then -- carriage return
					t &= #0D
				elsif c = 'n' then -- vertical line feed
					t &= #0A
				elsif c = 'x' then -- hex byte
					p += 1
					if p <= l then
						n = {line[p]}  p += 1
						if p <= l then
							n &= line[p]
						end if
						v = value('#' & upper(n))
						if v[1] != GET_FAIL then
							t &= v[2]
						end if
					end if   
				end if
			end if
		elsif c = ';' then
			terms &= {trim(t)}  t = {}
		else
			t &= c
		end if
		p += 1
	end while
	return terms
end function

function get_class_index(sequence id)
	for i = 1 to length(classes) do
		if equal(id,classes[i][classId]) then
			return i
		end if
	end for
	return 0
end function

function get_unit_index(sequence cid, sequence uid)
	for i = 1 to length(units) do
		if equal(units[i][unitId],uid) then
			if equal(units[i][unitClassId],cid) then
				return i
			end if
		end if
	end for
	return 0 
end function

global procedure load_conversions(sequence fname, sequence extra)
	object result
	integer fn
	sequence def,terms,id,class,from,to_,name,str
	fn = open(fname,"r")
	if fn != -1 then
		buf = {} -- clear input buffer
		while 1 do
			result = get_definition(fn)
			if atom(result) then
				exit
			end if
			def = result[1]
			terms = get_terms(result[2])
			id = {} class = {} from = {} to_ = {} name = {} str = {}
			for i = 1 to length(terms) do
				if match("id:",terms[i]) then
					id = trim(terms[i][4..length(terms[i])])
				elsif match("class:",terms[i]) then
					class = trim(terms[i][7..length(terms[i])])
				elsif match("from:",terms[i]) then
					from = trim(terms[i][6..length(terms[i])])
				elsif match("to:",terms[i]) then
					to_ = trim(terms[i][4..length(terms[i])])
				elsif match("name:",terms[i]) then
					name = trim(terms[i][6..length(terms[i])])
				elsif match("str:",terms[i]) then
					str = trim(terms[i][5..length(terms[i])])
				end if
			end for
			if equal(def,"class") then
				if length(id) and length(name) then
					if not get_class_index(id) then
						classes &= {{id,name & extra,str}}
					end if
				end if
			elsif equal(def,"unit") then
				if length(id) and length(class) and length(from) and length(to_) and length(name) then
					units &= {{id,class,from,to_,name,str}}
				end if
			elsif equal(def,"macro") then
				if length(id) and length(str) then
					units &= {{id,"m",str}}
				end if
			end if
		end while
		close(fn)
	end if
end procedure

global procedure empty_conversions()
	classes = {} units = {}
end procedure

global function get_class(sequence cid)
	integer i
	i = get_class_index(cid)
	if i then
		return classes[i]
	end if
	return {}
end function

global function get_unit(sequence cid, sequence uid)
	integer i
	i = get_unit_index(cid,uid)
	if i then
		return units[i]
	end if
	return {}
end function

integer elementIndex  elementIndex = 0
function sort_by_element(sequence a, sequence b)
	return compare(lower(a[elementIndex]),lower(b[elementIndex]))
end function

atom byElement  byElement = routine_id("sort_by_element")
function sort_ids_by_element(sequence s, integer index)
	elementIndex = index
	return custom_sort(byElement,s)
end function

global function get_class_ids(integer fsort)
	sequence c,t
	c = {}  t = classes
	if fsort then
		t = sort_ids_by_element(t,className)
	end if
	for i = 1 to length(t) do
		c &= {t[i][classId]}
	end for
	return c
end function


global function get_unit_ids_in_class(sequence cid, integer fsort)
	sequence u,t
	u = {}  t = units
	if fsort then
		t = sort_ids_by_element(t,unitName)
	end if
	for i = 1 to length(t) do
		if equal(t[i][unitClassId],cid) then
			u &= {t[i][unitId]}
		end if
	end for
	return u
end function

global function expand_macros(sequence s)
	sequence expd,m,t
	integer p,f,l,d,c,err
	expd = {}  err = 0
	while 1 do
		p = find(':',s)
		if not p then
			expd &= s  exit
		end if
		f = p-1
		while f > 0 do c = s[f]
			if c > ' ' then 
				f -= 1
			else
				exit
			end if
		end while
		l = p
		while l <= length(s) do c = s[l]
			if s[l] > ' ' then
				l += 1
			else
				exit
			end if
		end while
		expd &= s[1..f]  t = s[f+1..l-1]  s = s[l..length(s)]
		p = find(':',t) d = 0
		for i = length(t) to 1 by -1 do -- might be something like vol:gal.us.f
			if t[i] = '.' then  d = i exit  end if
		end for
		if not d then
			m = get_unit(t[1..p-1],t[p+1..length(t)])
			if not length(m) then err = 1 expd = t exit end if
			expd &= m[macroStr]
		else
			m = get_unit(t[1..p-1],t[p+1..d-1])
			if not length(m) then err = 1 expd = t exit end if
			if equal(t[d+1..length(t)],"t") then
				expd &= m[toUnit]					
			elsif equal(t[d+1..length(t)],"f") then
				expd &= m[fromUnit]
			else
				err = 1 expd = t exit
			end if
		end if
	end while
	return {err,expd}
end function

