
include misc.e
include get.e

sequence tokens, ids, safe, stack, buf
tokens = {}  ids = {}  safe = {}  stack = {}  buf = {}
sequence r   r = {0,0,0,0,0,0,0,0,0,0}
integer err  err = 0   sequence errstr  errstr = {}

procedure add(sequence name, integer id, integer narg)
	tokens &= {name} ids &= {id} safe &= narg
end procedure

procedure clr_err()
	err = 0  errstr = {}
end procedure

------------------------------------------------------------------
-- routines
------------------------------------------------------------------
procedure RESET()
	stack = {}
	r = {0,0,0,0,0,0,0,0,0,0}
	clr_err()
end procedure
add("reset",routine_id("RESET"),0)

integer i  i = 0
procedure STO()
	i = stack[1]
	if i > 0 and i <= 10 then
		r[i] = stack[2]
		stack = stack[3..length(stack)]
	else
		err = 1  errstr = "\'sto\' registers 1 thru 10 only "
	end if
end procedure
add("sto",routine_id("STO"),2)

procedure RCL()
	if stack[1] > 0 and stack[1] <= 10 then
		stack[1] = r[stack[1]]
	else
		err = 1  errstr = "\'rcl\' registers 1 thru 10 only "
	end if
end procedure
add("rcl",routine_id("RCL"),1)

procedure ADD()
	stack[2] += stack[1]
	stack = stack[2..length(stack)]
end procedure
add("+",routine_id("ADD"),2)

procedure SUB()
	stack[2] -= stack[1]
	stack = stack[2..length(stack)]
end procedure
add("-",routine_id("SUB"),2)

procedure MUL()
	stack[2] *= stack[1]
	stack = stack[2..length(stack)]
end procedure
add("*",routine_id("MUL"),2)

procedure DIV()
	stack[2] /= stack[1]
	stack = stack[2..length(stack)]
end procedure
add("/",routine_id("DIV"),2)

procedure REM()
	stack[2] = remainder(stack[2],stack[1])
	stack = stack[2..length(stack)]
end procedure
add("remainder",routine_id("REM"),2)

procedure PWR() -- forth F**
	stack[2] = power(stack[2],stack[1])
	stack = stack[2..length(stack)]
end procedure
add("^",routine_id("PWR"),2)
add("pwr",routine_id("PWR"),2) -- alias

procedure LN() -- natural logarithm
	stack[1] = log(stack[1]) -- antilog = e swap ^
end procedure
add("ln",routine_id("LN"),1)

procedure LOG() -- common logarithm
	stack[1] = log(stack[1]) / log(10) -- antilog = 10 swap pwr
end procedure
add("log",routine_id("LOG"),1)

procedure FLOOR()
	stack[1] = floor(stack[1])
end procedure
add("floor",routine_id("FLOOR"),1)

procedure SIGN()
	stack[1] = -(stack[1]<0)+(stack[1]>0)
end procedure
add("sign",routine_id("SIGN"),1)

procedure ABS()
	stack[1] *= -(stack[1]<0)+(stack[1]>0)
end procedure
add("abs",routine_id("ABS"),1)

atom p,x
procedure ROUND()
	p = power(10,stack[1])
	x = stack[2] stack[2] = floor((x*p)+0.5)/p
	stack = stack[2..length(stack)]
end procedure
add("round",routine_id("ROUND"),2)

procedure SIN()
	stack[1] = sin(stack[1])
end procedure
add("sin",routine_id("SIN"),1)

procedure COS()
	stack[1] = cos(stack[1])
end procedure
add("cos",routine_id("COS"),1)

procedure TAN()
	stack[1] = tan(stack[1])
end procedure
add("tan",routine_id("TAN"),1)

procedure ASIN()
	stack[1] = arcsin(stack[1])
end procedure
add("asin",routine_id("ASIN"),1)

procedure ACOS()
	stack[1] = arccos(stack[1])
end procedure
add("acos",routine_id("ACOS"),1)

procedure ATAN()
	stack[1] = arctan(stack[1])
end procedure
add("atan",routine_id("ATAN"),1)

procedure PI_()
	stack = prepend(stack,PI)
end procedure
add("pi",routine_id("PI_"),0)

procedure E()
	stack = prepend(stack,2.7182818284590452353602874713527)
end procedure
add("e",routine_id("E"),0)

procedure GRAV()-- standard acceleration of gravity meters per second per second
	stack = prepend(stack,9.80665) -- divide by 0.3048 for ft/s2 exact
end procedure
add("grav",routine_id("GRAV"),0)

procedure DROP()
	stack = stack[2..length(stack)]
end procedure
add("drop",routine_id("DROP"),1)

atom tmp
procedure SWAP()
	tmp = stack[1]
	stack[1] = stack[2]
	stack[2] = tmp
end procedure
add("swap",routine_id("SWAP"),2)

procedure DUP()
	stack = prepend(stack,stack[1])
end procedure
add("dup",routine_id("DUP"),1)

procedure NOP()
end procedure
add("nop",routine_id("NOP"),1)

procedure EXP()
	E() SWAP() PWR()
end procedure
add("exp",routine_id("EXP"),1)

------------------------------------------------------------------
-- parser
------------------------------------------------------------------
function token() -- parse a blank delimited word from input
	sequence s  s = {}
	for i = 1 to length(buf) do
		if buf[i] > 32 then
			buf = buf[i..length(buf)] exit
		end if
	end for
	for i = 1 to length(buf) do
		if buf[i] <= 32 then
			s = buf[1..i-1]
			buf = buf[i..length(buf)]
			return s
		end if
	end for
	s = buf  buf = {}
	return s
end function

function to_number(sequence s)
	for i = 1 to length(s) do
		if not find(s[i],"1234567890.EeABCDEF-+") then
			s = "" exit -- want 2drop and such to fail conversion
		end if
		if s[i] = '-' then
			if i > 1 then
				if i < length(s) then
					if s[i-1] != 'E' and s[i-1] != 'e' then
						s = "" exit
					end if
				else
					s = ""  -- want 1- to fail
				end if
			end if
		elsif s[i] = '+' then
			if i > 1 then
				if i < length(s) then
					if s[i-1] != 'E' and s[i-1] != 'e' then
						s = "" exit
					end if
				else
					s = ""
				end if
			end if
		end if
	end for
	return value(s)
end function

------------------------------------------------------------------
-- interpreter
------------------------------------------------------------------
function do_eval()
	sequence t,num
	integer p
	while 1 do
		t = token() num = {}
		if not length(t) then
			exit
		end if
		p = find(t,tokens)
		if p then
			if length(stack) >= safe[p] then
				call_proc(ids[p],{})
			else
				err = 1  errstr = "\'" & t & "\' needs " & sprint(safe[p]) & " on stack "
				exit
			end if
		else
			num = to_number(t)
			if not num[1] then
				stack = prepend(stack,num[2])
			else
				err = 1  errstr = "\'" & t & "\' is undefined "
				exit
			end if
		end if
	end while
	return {err,stack,errstr}
end function

global function eval(sequence s)
	buf = s  RESET()
	return do_eval()
end function

----------------------------------------------------
-- interactive interpreter
----------------------------------------------------
procedure BYE()
	abort(0)
end procedure
add("bye",routine_id("BYE"),0)

procedure DOTMEM() -- if called from a script this will open a console
	puts(1," ") ? r
end procedure
add(".mem",routine_id("DOTMEM"),0)

global procedure go()
	sequence  s
	puts(1,"Type \'bye\' to quit\nOK ")
	while 1 do
		clr_err()
		buf = gets(0)
		s = do_eval()
		puts(1,' ') ? stack
		if s[1] then
			stack = {}
			puts(1,s[3]) ? stack
		end if
		puts(1,"OK ")
	end while
end procedure

----------------------------------------------------

